@@ -62,6 +62,7 @@ typedef struct {
 
     TpHandleSet *tags;
 
+    GPtrArray *contact_info;
 } TpTestContactDetails;
 
 static TpTestContactDetails *
@@ -78,6 +79,9 @@ tp_test_contact_details_destroy (gpointer p)
   if (d->tags != NULL)
     tp_handle_set_destroy (d->tags);
 
+  if (d->contact_info != NULL)
+    g_ptr_array_unref (d->contact_info);
+
   g_free (d->id);
   g_free (d->alias);
   g_slice_free (TpTestContactDetails, d);
@@ -95,6 +99,7 @@ enum
 {
   ALIAS_UPDATED,
   PRESENCE_UPDATED,
+  CONTACT_INFO_UPDATED,
   N_SIGNALS
 };
 
@@ -374,6 +379,38 @@ static TpTestContactGroup *ensure_group (TpTestContactListManager *self,
 static TpTestContactList *ensure_list (TpTestContactListManager *self,
     TpTestContactListHandle handle);
 
+/*
+ * _insert_contact_field:
+ * @contact_info: an array of Contact_Info_Field structures
+ * @field_name: a vCard field name in any case combination
+ * @field_params: a list of vCard type-parameters, typically of the form
+ *  type=xxx; must be in lower-case if case-insensitive
+ * @field_values: for unstructured fields, an array containing one element;
+ *  for structured fields, the elements of the field in order
+ */
+static void
+_insert_contact_field (GPtrArray *contact_info,
+                       const gchar *field_name,
+                       const gchar * const *field_params,
+                       const gchar * const *field_values)
+{
+  const gchar * const *empty_strv = { NULL };
+  gchar *field_name_down = g_ascii_strdown (field_name, -1);
+
+  if (field_params == NULL)
+    field_params = empty_strv;
+  if (field_values == NULL)
+    field_values = empty_strv;
+
+   g_ptr_array_add (contact_info, tp_value_array_build (3,
+         G_TYPE_STRING, field_name_down,
+         G_TYPE_STRV, field_params,
+         G_TYPE_STRV, field_values,
+         G_TYPE_INVALID));
+
+   g_free (field_name_down);
+}
+
 static gboolean
 receive_contact_lists (gpointer p)
 {
@@ -432,6 +469,13 @@ receive_contact_lists (gpointer p)
   d->publish = TRUE;
   d->tags = tp_handle_set_new (self->priv->group_repo);
   tp_handle_set_add (d->tags, cambridge);
+  d->contact_info = dbus_g_type_specialized_construct (
+      TP_ARRAY_TYPE_CONTACT_INFO_FIELD_LIST);
+  {
+    const gchar * values[] = { "+14401223357708", NULL };
+    _insert_contact_field (d->contact_info, "tel", NULL,
+        (const gchar * const *) values);
+  }
   tp_handle_unref (self->priv->contact_repo, handle);
 
   id = "guillaume@example.com";
@@ -466,6 +510,33 @@ receive_contact_lists (gpointer p)
   d->tags = tp_handle_set_new (self->priv->group_repo);
   tp_handle_set_add (d->tags, montreal);
   tp_handle_set_add (d->tags, francophones);
+  d->contact_info = dbus_g_type_specialized_construct (
+      TP_ARRAY_TYPE_CONTACT_INFO_FIELD_LIST);
+  {
+    const gchar * values[] = { "+15142345678", NULL };
+    _insert_contact_field (d->contact_info, "tel", NULL,
+        (const gchar * const *) values);
+  }
+  {
+    const gchar * values[] = { "1982-01-02T13:57Z", NULL };
+    _insert_contact_field (d->contact_info, "tel", NULL,
+        (const gchar * const *) values);
+  }
+  {
+    const gchar * values[] = { "Olivier Crete", NULL };
+    _insert_contact_field (d->contact_info, "fn", NULL,
+        (const gchar * const *) values);
+  }
+  {
+    const gchar * values[] = { id, NULL };
+    _insert_contact_field (d->contact_info, "email", NULL,
+        (const gchar * const *) values);
+  }
+  {
+    const gchar * values[] = { "ocrete.example.com", NULL };
+    _insert_contact_field (d->contact_info, "url", NULL,
+        (const gchar * const *) values);
+  }
   tp_handle_unref (self->priv->contact_repo, handle);
 
   id = "travis@example.com";
@@ -496,6 +567,7 @@ receive_contact_lists (gpointer p)
     {
       g_signal_emit (self, signals[ALIAS_UPDATED], 0, handle);
       g_signal_emit (self, signals[PRESENCE_UPDATED], 0, handle);
+      g_signal_emit (self, signals[CONTACT_INFO_UPDATED], 0, handle);
     }
 
   tp_intset_destroy (set);
@@ -548,6 +620,7 @@ receive_contact_lists (gpointer p)
     {
       g_signal_emit (self, signals[ALIAS_UPDATED], 0, handle);
       g_signal_emit (self, signals[PRESENCE_UPDATED], 0, handle);
+      g_signal_emit (self, signals[CONTACT_INFO_UPDATED], 0, handle);
     }
 
   tp_intset_destroy (set);
@@ -576,6 +649,7 @@ receive_contact_lists (gpointer p)
   tp_intset_destroy (set);
   g_signal_emit (self, signals[ALIAS_UPDATED], 0, handle);
   g_signal_emit (self, signals[PRESENCE_UPDATED], 0, handle);
+  g_signal_emit (self, signals[CONTACT_INFO_UPDATED], 0, handle);
 
   id = "christian@example.com";
   handle = tp_handle_ensure (self->priv->contact_repo, id, NULL, NULL);
@@ -598,6 +672,7 @@ receive_contact_lists (gpointer p)
   tp_intset_destroy (set);
   g_signal_emit (self, signals[ALIAS_UPDATED], 0, handle);
   g_signal_emit (self, signals[PRESENCE_UPDATED], 0, handle);
+  g_signal_emit (self, signals[CONTACT_INFO_UPDATED], 0, handle);
 
   tp_group_mixin_change_members ((GObject *) cambridge_group, "",
       cam_set, NULL, NULL, NULL,
@@ -724,6 +799,13 @@ tp_test_contact_list_manager_class_init (TpTestContactListManagerClass *klass)
       0,
       NULL, NULL,
       g_cclosure_marshal_VOID__UINT, G_TYPE_NONE, 1, G_TYPE_UINT);
+
+  signals[CONTACT_INFO_UPDATED] = g_signal_new ("contact-info-updated",
+      G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST,
+      0,
+      NULL, NULL,
+      g_cclosure_marshal_VOID__UINT, G_TYPE_NONE, 1, G_TYPE_UINT);
 }
 
 static void
@@ -1070,12 +1152,53 @@ send_updated_roster (TpTestContactListManager *self,
             }
         }
 
+      if (d->contact_info == NULL)
+        {
+          g_message ("\tno contact info");
+        }
+      else
+        {
+          for (guint i = 0; i < d->contact_info->len; i++)
+            {
+              GValueArray *va = g_ptr_array_index (d->contact_info, i);
+
+              GValue *gv0, *gv1, *gv2;
+              gchar **params, **values;
+              gchar *empty_strv[] = { NULL };
+
+              gv0 = g_value_array_get_nth (va, 0);
+              gv1 = g_value_array_get_nth (va, 1);
+              gv2 = g_value_array_get_nth (va, 2);
+              params = g_value_peek_pointer (gv1);
+              values = g_value_peek_pointer (gv2);
+
+              if (params == NULL)
+                params = empty_strv;
+              if (values == NULL)
+                values = empty_strv;
+
+              g_message ("\tcontact info:");
+              g_message ("\t\tname: %s", g_value_get_string (gv0));
+              g_message ("\t\tparams:");
+              for (guint j = 0; j < g_strv_length (params); j++)
+                {
+                  g_message ("\t\t\t%s", params[j]);
+                }
+              g_message ("\t\tvalues:");
+              for (guint j = 0; j < g_strv_length (values); j++)
+                {
+                  g_message ("\t\t\t%s", values[j]);
+                }
+            }
+        }
+
       handle = tp_handle_ensure (self->priv->contact_repo, d->id, NULL, NULL);
       /* XXX: ideally, we'd only do this if thse specific details really
        * changed, but it's not terribly unrealistic to think some servers or CMs
        * would make this mistake as well. */
       g_signal_emit (self, signals[ALIAS_UPDATED], 0, handle);
       g_signal_emit (self, signals[PRESENCE_UPDATED], 0, handle);
+      g_signal_emit (self, signals[CONTACT_INFO_UPDATED], 0, handle);
 
       tp_handle_unref (self->priv->contact_repo, handle);
     }
@@ -1683,3 +1806,63 @@ tp_test_contact_list_manager_set_alias (TpTestContactListManager *self,
       TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
   tp_intset_destroy (set);
 }
+
+GPtrArray *
+tp_test_contact_list_manager_get_contact_info (TpTestContactListManager *self,
+                                               TpHandle contact)
+{
+  TpTestContactDetails *d = lookup_contact (self, contact);
+
+  if (d != NULL)
+    return d->contact_info;
+
+  return NULL;
+}
+
+void
+tp_test_contact_list_manager_set_contact_info (TpTestContactListManager *self,
+                                               const GPtrArray *contact_info)
+{
+  TpTestContactList *stored = self->priv->lists[
+    TP_TEST_CONTACT_LIST_STORED];
+  GPtrArray *old;
+  TpTestContactDetails *d = ensure_contact (self, self->priv->conn->self_handle,
+      NULL);
+  d->id = g_strdup (tp_handle_inspect (self->priv->contact_repo,
+        self->priv->conn->self_handle));
+  old = d->contact_info;
+
+  /* FIXME: if stored list hasn't been retrieved yet, queue the change for
+   * later */
+
+  /* if shutting down, do nothing */
+  if (stored == NULL)
+    return;
+
+  d->contact_info = dbus_g_type_specialized_construct (
+      TP_ARRAY_TYPE_CONTACT_INFO_FIELD_LIST);
+  {
+    guint i;
+    for (i = 0; i < contact_info->len; i++)
+      {
+        const gchar *name;
+        const gchar * const * params;
+        const gchar * const * values;
+        GValueArray *va = g_ptr_array_index (contact_info, i);
+
+        tp_value_array_unpack (va, 3,
+          &name,
+          &params,
+          &values);
+
+        _insert_contact_field (d->contact_info, name, params, values);
+      }
+  }
+
+  /* always send the updated roster, since it's not worth checking the
+   * contact_info for changes */
+  send_updated_roster (self, self->priv->conn->self_handle);
+
+  if (old != NULL)
+    g_ptr_array_unref (old);
+}
